@@ -5,7 +5,8 @@
  */
 
 #include "doctor.h"
-#include "dpmsclient.h"
+#include "mode.h"
+#include <dpms.h>
 
 #include <QCollator>
 #include <QCoreApplication>
@@ -17,11 +18,14 @@
 #include <QJsonObject>
 #include <QLoggingCategory>
 #include <QRect>
+#include <QScreen>
 #include <QStandardPaths>
 
+#include <utility>
+
 #include "../backendmanager_p.h"
+#include "../config.h"
 #include "../configoperation.h"
-#include "../edid.h"
 #include "../getconfigoperation.h"
 #include "../log.h"
 #include "../output.h"
@@ -83,19 +87,45 @@ void Doctor::start(QCommandLineParser *parser)
             return;
         }
 
-        m_dpmsClient = new DpmsClient(this);
+        m_dpmsClient = new Dpms(this);
+        auto screens = qGuiApp->screens();
         if (m_parser->isSet(QStringLiteral("dpms-excluded"))) {
             const auto excludedConnectors = m_parser->values(QStringLiteral("dpms-excluded"));
-            m_dpmsClient->setExcludedOutputNames(excludedConnectors);
+            auto it = std::remove_if(screens.begin(), screens.end(), [&excludedConnectors](QScreen *screen) {
+                return excludedConnectors.contains(screen->name());
+            });
+            screens.erase(it, screens.end());
         }
 
-        connect(m_dpmsClient, &DpmsClient::finished, qApp, &QCoreApplication::quit);
+        connect(m_dpmsClient, &Dpms::hasPendingChangesChanged, qGuiApp, [](bool hasChanges) {
+            if (!hasChanges) {
+                qGuiApp->quit();
+            }
+        });
 
         const QString dpmsArg = m_parser->value(QStringLiteral("dpms"));
         if (dpmsArg == QLatin1String("show")) {
-            showDpms();
         } else {
-            setDpms(dpmsArg);
+            auto performSwitch = [this, dpmsArg, screens](bool supported) {
+                if (!supported) {
+                    cerr << "DPMS not supported in this system";
+                    qGuiApp->quit();
+                    return;
+                }
+
+                if (dpmsArg == QLatin1String("off")) {
+                    m_dpmsClient->switchMode(KScreen::Dpms::Off, screens);
+                } else if (dpmsArg == QLatin1String("on")) {
+                    m_dpmsClient->switchMode(KScreen::Dpms::On, screens);
+                } else {
+                    cerr << "--dpms argument not understood (" << dpmsArg << ")";
+                }
+            };
+            if (m_dpmsClient->isSupported()) {
+                performSwitch(m_dpmsClient->isSupported());
+            } else {
+                connect(m_dpmsClient, &Dpms::supportedChanged, this, performSwitch);
+            }
         }
         return;
     }
@@ -112,34 +142,6 @@ void Doctor::start(QCommandLineParser *parser)
     QTimer::singleShot(0, qApp->quit);
 }
 
-void KScreen::Doctor::setDpms(const QString &dpmsArg)
-{
-    qDebug() << "SetDpms: " << dpmsArg;
-    connect(m_dpmsClient, &DpmsClient::ready, this, [this, dpmsArg]() {
-        cout << "DPMS.ready()";
-        if (dpmsArg == QLatin1String("off")) {
-            m_dpmsClient->off();
-        } else if (dpmsArg == QLatin1String("on")) {
-            m_dpmsClient->on();
-        } else {
-            cout << "--dpms argument not understood (" << dpmsArg << ")";
-        }
-    });
-
-    m_dpmsClient->connect();
-}
-
-void Doctor::showDpms()
-{
-    m_dpmsClient = new DpmsClient(this);
-
-    connect(m_dpmsClient, &DpmsClient::ready, this, []() {
-        cout << "DPMS.ready()";
-    });
-
-    m_dpmsClient->connect();
-}
-
 void Doctor::showBackends() const
 {
     cout << "Environment: " << Qt::endl;
@@ -170,53 +172,70 @@ void Doctor::setOptionList(const QStringList &outputArgs)
     m_outputArgs = outputArgs;
 }
 
+OutputPtr Doctor::findOutput(const QString &query)
+{
+    // try as an output name or ID
+    for (const auto &output : m_config->outputs()) {
+        if (output->name() == query) {
+            return output;
+        }
+    }
+    bool ok;
+    int id = query.toInt(&ok);
+    if (!ok) {
+        cerr << "Output with name " << query << " not found." << Qt::endl;
+        return OutputPtr();
+    }
+
+    if (m_config->outputs().contains(id)) {
+        return m_config->outputs()[id];
+    } else {
+        cerr << "Output with id " << id << " not found." << Qt::endl;
+        return OutputPtr();
+    }
+}
+
 void Doctor::parseOutputArgs()
 {
     // qCDebug(KSCREEN_DOCTOR) << "POSARGS" << m_positionalArgs;
-    for (const QString &op : qAsConst(m_outputArgs)) {
+    for (const QString &op : std::as_const(m_outputArgs)) {
         auto ops = op.split(QLatin1Char('.'));
         if (ops.count() > 2) {
             bool ok;
-            int output_id = -1;
             if (ops[0] == QLatin1String("output")) {
-                for (const auto &output : m_config->outputs()) {
-                    if (output->name() == ops[1]) {
-                        output_id = output->id();
-                    }
+                OutputPtr output = findOutput(ops[1]);
+                if (!output) {
+                    qApp->exit(3);
+                    return;
                 }
-                if (output_id == -1) {
-                    output_id = ops[1].toInt(&ok);
-                    if (!ok) {
-                        cerr << "Unable to parse output id: " << ops[1] << Qt::endl;
-                        qApp->exit(3);
+                int output_id = output->id();
+
+                const QString subcmd = ops.length() > 2 ? ops[2] : QString();
+
+                if (ops.count() == 3 && subcmd == QLatin1String("primary")) {
+                    setPrimary(output);
+                } else if (ops.count() == 4 && subcmd == QLatin1String("priority")) {
+                    uint32_t priority = ops[3].toUInt(&ok);
+                    if (!ok || priority > 100) {
+                        qCWarning(KSCREEN_DOCTOR) << "Wrong input: allowed values for priority are from 1 to 100";
+                        qApp->exit(5);
                         return;
                     }
-                }
-                if (ops.count() == 3 && ops[2] == QLatin1String("primary")) {
-                    if (!setPrimary(output_id)) {
-                        qApp->exit(1);
-                        return;
-                    };
-                } else if (ops.count() == 3 && ops[2] == QLatin1String("enable")) {
-                    if (!setEnabled(output_id, true)) {
-                        qApp->exit(1);
-                        return;
-                    };
-                } else if (ops.count() == 3 && ops[2] == QLatin1String("disable")) {
-                    if (!setEnabled(output_id, false)) {
-                        qApp->exit(1);
-                        return;
-                    };
-                } else if (ops.count() == 4 && ops[2] == QLatin1String("mode")) {
+                    setPriority(output, priority);
+                } else if (ops.count() == 3 && subcmd == QLatin1String("enable")) {
+                    setEnabled(output, true);
+                } else if (ops.count() == 3 && subcmd == QLatin1String("disable")) {
+                    setEnabled(output, false);
+                } else if (ops.count() == 4 && subcmd == QLatin1String("mode")) {
                     QString mode_id = ops[3];
                     // set mode
-                    if (!setMode(output_id, mode_id)) {
+                    if (!setMode(output, mode_id)) {
                         qApp->exit(9);
                         return;
                     }
                     qCDebug(KSCREEN_DOCTOR) << "Output" << output_id << "set mode" << mode_id;
 
-                } else if (ops.count() == 4 && ops[2] == QLatin1String("position")) {
+                } else if (ops.count() == 4 && subcmd == QLatin1String("position")) {
                     QStringList _pos = ops[3].split(QLatin1Char(','));
                     if (_pos.count() != 2) {
                         qCWarning(KSCREEN_DOCTOR) << "Invalid position:" << ops[3];
@@ -233,11 +252,9 @@ void Doctor::parseOutputArgs()
 
                     QPoint p(x, y);
                     qCDebug(KSCREEN_DOCTOR) << "Output position" << p;
-                    if (!setPosition(output_id, p)) {
-                        qApp->exit(1);
-                        return;
-                    }
-                } else if ((ops.count() == 4 || ops.count() == 5) && ops[2] == QLatin1String("scale")) {
+                    setPosition(output, p);
+
+                } else if ((ops.count() == 4 || ops.count() == 5) && subcmd == QLatin1String("scale")) {
                     // be lenient about . vs. comma as separator
                     qreal scale = ops[3].replace(QLatin1Char(','), QLatin1Char('.')).toDouble(&ok);
                     if (ops.count() == 5) {
@@ -245,12 +262,13 @@ void Doctor::parseOutputArgs()
                         scale = dbl.toDouble(&ok);
                     };
                     // set scale
-                    if (!ok || qFuzzyCompare(scale, 0.0) || !setScale(output_id, scale)) {
+                    if (!ok || qFuzzyCompare(scale, 0.0)) {
                         qCDebug(KSCREEN_DOCTOR) << "Could not set scale " << scale << " to output " << output_id;
                         qApp->exit(9);
                         return;
                     }
-                } else if ((ops.count() == 4) && (ops[2] == QLatin1String("orientation") || ops[2] == QStringLiteral("rotation"))) {
+                    setScale(output, scale);
+                } else if ((ops.count() == 4) && (subcmd == QLatin1String("orientation") || subcmd == QStringLiteral("rotation"))) {
                     const QString _rotation = ops[3].toLower();
                     bool ok = false;
                     const QHash<QString, KScreen::Output::Rotation> rotationMap({{QStringLiteral("none"), KScreen::Output::None},
@@ -264,24 +282,21 @@ void Doctor::parseOutputArgs()
                         ok = true;
                         rot = rotationMap[_rotation];
                     }
-                    if (!ok || !setRotation(output_id, rot)) {
+                    if (!ok) {
                         qCDebug(KSCREEN_DOCTOR) << "Could not set orientation " << _rotation << " to output " << output_id;
                         qApp->exit(9);
                         return;
                     }
-                } else if (ops.count() == 4 && ops[2] == QLatin1String("overscan")) {
+                    setRotation(output, rot);
+                } else if (ops.count() == 4 && subcmd == QLatin1String("overscan")) {
                     const uint32_t overscan = ops[3].toInt();
                     if (overscan > 100) {
                         qCWarning(KSCREEN_DOCTOR) << "Wrong input: allowed values for overscan are from 0 to 100";
                         qApp->exit(9);
                         return;
                     }
-                    if (!setOverscan(output_id, overscan)) {
-                        qCDebug(KSCREEN_DOCTOR) << "Could not set overscan " << overscan << " to output " << output_id;
-                        qApp->exit(9);
-                        return;
-                    }
-                } else if (ops.count() == 4 && ops[2] == QLatin1String("vrrpolicy")) {
+                    setOverscan(output, overscan);
+                } else if (ops.count() == 4 && subcmd == QLatin1String("vrrpolicy")) {
                     const QString _policy = ops[3].toLower();
                     KScreen::Output::VrrPolicy policy;
                     if (_policy == QStringLiteral("never")) {
@@ -295,12 +310,8 @@ void Doctor::parseOutputArgs()
                         qApp->exit(9);
                         return;
                     }
-                    if (!setVrrPolicy(output_id, policy)) {
-                        qCDebug(KSCREEN_DOCTOR) << "Could not set vrr policy " << policy << " to output " << output_id;
-                        qApp->exit(9);
-                        return;
-                    }
-                } else if (ops.count() == 4 && ops[2] == QLatin1String("rgbrange")) {
+                    setVrrPolicy(output, policy);
+                } else if (ops.count() == 4 && subcmd == QLatin1String("rgbrange")) {
                     const QString _range = ops[3].toLower();
                     KScreen::Output::RgbRange range;
                     if (_range == QStringLiteral("automatic")) {
@@ -314,11 +325,7 @@ void Doctor::parseOutputArgs()
                         qApp->exit(9);
                         return;
                     }
-                    if (!setRgbRange(output_id, range)) {
-                        qCDebug(KSCREEN_DOCTOR) << "Could not set rgb range " << range << " to output " << output_id;
-                        qApp->exit(9);
-                        return;
-                    }
+                    setRgbRange(output, range);
                 } else {
                     cerr << "Unable to parse arguments: " << op << Qt::endl;
                     qApp->exit(2);
@@ -333,6 +340,11 @@ void Doctor::configReceived(KScreen::ConfigOperation *op)
 {
     m_config = op->config();
 
+    if (!m_config) {
+        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
+        return;
+    }
+
     if (m_parser->isSet(QStringLiteral("json"))) {
         showJson();
         qApp->quit();
@@ -350,22 +362,8 @@ void Doctor::configReceived(KScreen::ConfigOperation *op)
     }
 }
 
-int Doctor::outputCount() const
-{
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return 0;
-    }
-    return m_config->outputs().count();
-}
-
 void Doctor::showOutputs() const
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return;
-    }
-
     QHash<KScreen::Output::Type, QString> typeString;
     typeString[KScreen::Output::Unknown] = QStringLiteral("Unknown");
     typeString[KScreen::Output::VGA] = QStringLiteral("VGA");
@@ -388,9 +386,9 @@ void Doctor::showOutputs() const
 
     for (const auto &output : m_config->outputs()) {
         cout << green << "Output: " << cr << output->id() << " " << output->name();
-        cout << " " << (output->isEnabled() ? green + QStringLiteral("enabled") : red + QStringLiteral("disabled"));
-        cout << " " << (output->isConnected() ? green + QStringLiteral("connected") : red + QStringLiteral("disconnected"));
-        cout << " " << (output->isPrimary() ? green + QStringLiteral("primary") : QString());
+        cout << " " << (output->isEnabled() ? green + QStringLiteral("enabled") : red + QStringLiteral("disabled")) << cr;
+        cout << " " << (output->isConnected() ? green + QStringLiteral("connected") : red + QStringLiteral("disconnected")) << cr;
+        cout << " " << (output->isEnabled() ? green : red) + QStringLiteral("priority ") << output->priority() << cr;
         auto _type = typeString[output->type()];
         cout << " " << yellow << (_type.isEmpty() ? QStringLiteral("UnmappedOutputType") : _type);
         cout << blue << " Modes: " << cr;
@@ -436,19 +434,16 @@ void Doctor::showOutputs() const
         if (output->capabilities() & Output::Capability::RgbRange) {
             switch (output->rgbRange()) {
             case Output::RgbRange::Automatic:
-                cout << cr << "Automatic ";
+                cout << cr << "Automatic";
                 break;
             case Output::RgbRange::Full:
-                cout << cr << "Full ";
+                cout << cr << "Full";
                 break;
             case Output::RgbRange::Limited:
-                cout << cr << "Limited ";
+                cout << cr << "Limited";
             }
         } else {
-            cout << cr << "unknown ";
-        }
-        if (output->isPrimary()) {
-            cout << blue << "primary";
+            cout << cr << "unknown";
         }
         cout << cr << Qt::endl;
     }
@@ -460,186 +455,85 @@ void Doctor::showJson() const
     cout << doc.toJson(QJsonDocument::Indented);
 }
 
-bool Doctor::setEnabled(int id, bool enabled = true)
+void Doctor::setEnabled(OutputPtr output, bool enable)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return false;
-    }
-
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            cout << (enabled ? "Enabling " : "Disabling ") << "output " << id << Qt::endl;
-            output->setEnabled(enabled);
-            m_changed = true;
-            return true;
-        }
-    }
-    cerr << "Output with id " << id << " not found." << Qt::endl;
-    qApp->exit(8);
-    return false;
+    cout << (enable ? "Enabling " : "Disabling ") << "output " << output->id() << Qt::endl;
+    output->setEnabled(enable);
+    m_changed = true;
 }
 
-bool Doctor::setPosition(int id, const QPoint &pos)
+void Doctor::setPosition(OutputPtr output, const QPoint &pos)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return false;
-    }
-
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            qCDebug(KSCREEN_DOCTOR) << "Set output position" << pos;
-            output->setPos(pos);
-            m_changed = true;
-            return true;
-        }
-    }
-    cout << "Output with id " << id << " not found." << Qt::endl;
-    return false;
+    qCDebug(KSCREEN_DOCTOR) << "Set output position" << pos;
+    output->setPos(pos);
+    m_changed = true;
 }
 
-bool Doctor::setMode(int id, const QString &mode_id)
+KScreen::ModePtr Doctor::findMode(OutputPtr output, const QString &query)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return false;
-    }
-
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            // find mode
-            for (const KScreen::ModePtr &mode : output->modes()) {
-                auto name =
-                    QStringLiteral("%1x%2@%3")
+    for (const KScreen::ModePtr &mode : output->modes()) {
+        auto name = QStringLiteral("%1x%2@%3")
                         .arg(QString::number(mode->size().width()), QString::number(mode->size().height()), QString::number(qRound(mode->refreshRate())));
-                if (mode->id() == mode_id || name == mode_id) {
-                    qCDebug(KSCREEN_DOCTOR) << "Taddaaa! Found mode" << mode->id() << name;
-                    output->setCurrentModeId(mode->id());
-                    m_changed = true;
-                    return true;
-                }
-            }
+        if (mode->id() == query || name == query) {
+            qCDebug(KSCREEN_DOCTOR) << "Taddaaa! Found mode" << mode->id() << name;
+            return mode;
         }
     }
-    cout << "Output mode " << mode_id << " not found." << Qt::endl;
-    return false;
+    cout << "Output mode " << query << " not found." << Qt::endl;
+    return ModePtr();
 }
 
-bool Doctor::setScale(int id, qreal scale)
+bool Doctor::setMode(OutputPtr output, const QString &query)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
+    // find mode
+    const KScreen::ModePtr mode = findMode(output, query);
+    if (!mode) {
         return false;
     }
-
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            output->setScale(scale);
-            m_changed = true;
-            return true;
-        }
-    }
-    cout << "Output scale " << id << " invalid." << Qt::endl;
-    return false;
+    output->setCurrentModeId(mode->id());
+    m_changed = true;
+    return true;
 }
 
-bool Doctor::setRotation(int id, KScreen::Output::Rotation rot)
+void Doctor::setScale(OutputPtr output, qreal scale)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return false;
-    }
-
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            output->setRotation(rot);
-            m_changed = true;
-            return true;
-        }
-    }
-    cout << "Output rotation " << id << " invalid." << Qt::endl;
-    return false;
+    output->setScale(scale);
+    m_changed = true;
 }
 
-bool Doctor::setOverscan(int id, uint32_t overscan)
+void Doctor::setRotation(OutputPtr output, KScreen::Output::Rotation rot)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return false;
-    }
-
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            output->setOverscan(overscan);
-            m_changed = true;
-            return true;
-        }
-    }
-    cout << "Output overscan " << id << " invalid." << Qt::endl;
-    return false;
+    output->setRotation(rot);
+    m_changed = true;
 }
 
-bool Doctor::setVrrPolicy(int id, KScreen::Output::VrrPolicy policy)
+void Doctor::setOverscan(OutputPtr output, uint32_t overscan)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return false;
-    }
-
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            output->setVrrPolicy(policy);
-            m_changed = true;
-            return true;
-        }
-    }
-    cout << "Output VrrPolicy " << id << " invalid." << Qt::endl;
-    return false;
+    output->setOverscan(overscan);
+    m_changed = true;
 }
 
-bool Doctor::setRgbRange(int id, KScreen::Output::RgbRange rgbRange)
+void Doctor::setVrrPolicy(OutputPtr output, KScreen::Output::VrrPolicy policy)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return false;
-    }
-
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            output->setRgbRange(rgbRange);
-            m_changed = true;
-            return true;
-        }
-    }
-    cout << "Output RgbRange " << id << " invalid." << Qt::endl;
-    return false;
+    output->setVrrPolicy(policy);
+    m_changed = true;
 }
 
-bool KScreen::Doctor::setPrimary(int id)
+void Doctor::setRgbRange(OutputPtr output, KScreen::Output::RgbRange rgbRange)
 {
-    if (!m_config) {
-        qCWarning(KSCREEN_DOCTOR) << "Invalid config.";
-        return false;
-    }
+    output->setRgbRange(rgbRange);
+    m_changed = true;
+}
 
-    bool found = false;
-    for (const auto &output : m_config->outputs()) {
-        if (output->id() == id) {
-            output->setPrimary(true);
-            m_changed = true;
-            found = true;
-        }
-    }
-    if (found) {
-        for (const auto &output : m_config->outputs()) {
-            if (output->id() != id) {
-                output->setPrimary(false);
-                m_changed = true;
-            }
-        }
-    }
-    return found;
+void KScreen::Doctor::setPrimary(OutputPtr output)
+{
+    setPriority(output, 1);
+}
+
+void KScreen::Doctor::setPriority(OutputPtr output, uint32_t priority)
+{
+    m_config->setOutputPriority(output, priority);
+    m_changed = true;
 }
 
 void Doctor::applyConfig()
